include	<ctype.h>
include	<imhdr.h>
include	"ace.h"


# CONVOLVES -- Get a line of convolved data.  Also get the unconvolved
# data, the sky data, sigma data, and exposure data.
#
# This routine must be called sequentially starting with the first line.
# It is initialized when the first line.  Memory is freed by using a final
# call with a line of zero.

procedure convolves (in, bpm, sky, sig, exp, bpval, bpdet, bpflg, offset, scale,
	line, cnv, indata, bp, cnvdata, scnvdata, skydata, sigdata, expdata,
	cnvwt, logfd)

pointer	in[2]		#I Image pointers
pointer	bpm[2]		#I BPM pointer
pointer	sky[2]		#I Sky map
pointer	sig[2]		#I Sigma map
pointer	exp[2]		#I Exposure map
int	bpval		#I Output bad pixel value
int	bpdet		#I Detection bad pixel ranges
int	bpflg		#I Flag bad pixel ranges
int	offset[2]	#I Offsets
real	scale[2]	#I Image scales
int	line		#I Line
char	cnv[ARB]	#I Convolution string
pointer	indata[2]	#O Pointers to unconvolved image data
pointer	bp		#O Bad pixel data
pointer	cnvdata		#O Pointer to convolved image data
pointer	scnvdata	#O Pointer to convolved sky data
pointer	skydata[2]	#O Pointer to sky data
pointer	sigdata[2]	#O Pointer to sigma data corrected by exposure map
pointer	expdata[2]	#O Pointer to exposure map data
real	cnvwt		#O Weight for convolved sigma
int	logfd		#I Logfile

int	i, j, k,  nx, ny, nx2, ny2, nc, nl, mode, off
real	wts, wts1
pointer	bpm2, kptr, ptr, sptr, symptr, symwptr
bool	dobpm, overlap, fp_equalr()

pointer	kernel, sym, symbuf, symwts
pointer bpbuf, bpwts, wtsl, scales
pointer	buf, buf2, buf3
pointer	sbuf, sbuf2, sbuf3
pointer	sigsbuf, sigsbuf2
pointer	ebuf, ebuf2
data	kernel/NULL/, sym/NULL/, symbuf/NULL/, symwts/NULL/
data	bpbuf/NULL/, bpwts/NULL/, wtsl/NULL/, scales/NULL/
data	buf/NULL/, buf2/NULL/, buf3/NULL/
data	sbuf/NULL/, sbuf2/NULL/, sbuf3/NULL/
data	sigbuf/NULL/, sigbuf2/NULL/
data	ebuf/NULL/, ebuf2/NULL/

errchk	cnvparse, cnvgline2

begin
	# If no convolution.
	if (cnv[1] == EOS) {
	    if (line == 0)
		return

	    call cnvgline1 (line, offset, in, bpm, indata,
	        bp, bpval, bpdet, bpflg)
	    call cnvgline2 (line, offset, in, sky, sig, exp, skydata,
		sigdata, expdata)
	    cnvwt = 1
	    if (in[2] == NULL) {
		cnvdata = indata[1]
		scnvdata = skydata[1]
	    } else {
		call asubr_scale (Memr[indata[1]], scale[1],
		    Memr[indata[2]], scale[2], Memr[cnvdata], IM_LEN(in[1],1))
		call asubr_scale (Memr[skydata[1]], scale[1],
		    Memr[skydata[2]], scale[2], Memr[scnvdata], IM_LEN(in[1],1))
	    }
	    return
	}

	# Free memory.
	if (line == 0) {
	    if (symbuf != NULL) {
		do i = 0, ARB {
		    ptr = Memi[symbuf+i]
		    if (ptr == -1)
			break
		    call mfree (ptr, TY_REAL)
		}
	    }
	    if (symwts != NULL) {
		do i = 0, ARB {
		    ptr = Memi[symwts+i]
		    if (ptr == -1)
			break
		    call mfree (ptr, TY_REAL)
		}
	    }
	    call mfree (scales, TY_REAL)
	    call mfree (wtsl, TY_REAL)
	    call mfree (kernel, TY_REAL)
	    call mfree (scales, TY_REAL)
	    call mfree (sym, TY_INT)
	    call mfree (symbuf, TY_POINTER)
	    call mfree (symwts, TY_POINTER)
	    call mfree (buf, TY_REAL)
	    call mfree (buf2, TY_REAL)
	    call mfree (buf3, TY_REAL)
	    call mfree (bpbuf, TY_REAL)
	    call mfree (bpwts, TY_REAL)
	    call mfree (sbuf, TY_REAL)
	    call mfree (sbuf2, TY_REAL)
	    call mfree (sbuf3, TY_REAL)
	    call mfree (sigbuf, TY_REAL)
	    call mfree (ebuf, TY_REAL)

	    return
	}

	# Initialize by getting the kernel coefficients, setting the
	# image I/O buffers using a scrolling array, and allocate memory.

	if (line == 1 || buf == NULL) {
	    if (buf != NULL) {
		if (symbuf != NULL) {
		    do i = 0, ARB {
			ptr = Memi[symbuf+i]
			if (ptr == -1)
			    break
			call mfree (ptr, TY_REAL)
		    }
		}
		if (symwts != NULL) {
		    do i = 0, ARB {
			ptr = Memi[symwts+i]
			if (ptr == -1)
			    break
			call mfree (ptr, TY_REAL)
		    }
		}
		call mfree (scales, TY_REAL)
		call mfree (wtsl, TY_REAL)
		call mfree (kernel, TY_REAL)
		call mfree (scales, TY_REAL)
		call mfree (sym, TY_INT)
		call mfree (symbuf, TY_POINTER)
		call mfree (symwts, TY_POINTER)
		call mfree (buf, TY_REAL)
		call mfree (buf2, TY_REAL)
		call mfree (buf3, TY_REAL)
		call mfree (bpbuf, TY_REAL)
		call mfree (bpwts, TY_REAL)
		call mfree (sbuf, TY_REAL)
		call mfree (sbuf2, TY_REAL)
		call mfree (sbuf3, TY_REAL)
		call mfree (sigbuf, TY_REAL)
		call mfree (ebuf, TY_REAL)
	    }

	    nc = IM_LEN(in[1],1)
	    nl = IM_LEN(in[1],2)

	    call cnvparse (cnv, kernel, nx, ny, logfd)
	    nx2 = nx / 2
	    ny2 = ny / 2
	    call malloc (scales, ny, TY_REAL)
	    call calloc (wtsl, ny, TY_REAL)
	    call amovkr (1., Memr[scales], ny)

	    # Check for lines which are simple scalings of the first line.
	    do i = 2, ny {
		kptr = kernel + (i - 1) * nx
		wts = 0.
		do k = 0, nx-1 {
		    if (Memr[kptr+k] == 0. || Memr[kernel+k] == 0.) {
			wts = 0.
			break
		    }
		    if (wts == 0.)
			wts = Memr[kptr+k] / Memr[kernel+k]
		    else {
			wts1 = Memr[kptr+k] / Memr[kernel+k]
			if (!fp_equalr (wts, wts1))
			    break
		    }
		}
		if (wts != 0. && fp_equalr (wts, wts1)) {
		    Memr[scales+i-1] = wts
		    call amovr (Memr[kernel], Memr[kptr], nx)
		}
	    }

	    wts = 0
	    do i = 1, ny {
		kptr = kernel + (i - 1) * nx
		wts1 = 0.
		do j = 1, nx {
		    wts1 = wts1 + Memr[kptr]
		    kptr = kptr + 1
		}
		Memr[wtsl+i-1] = wts1
		wts = wts + wts1
	    }
	    if (wts != 0.) {
		call adivkr (Memr[wtsl], wts, Memr[wtsl], ny)
		call adivkr (Memr[kernel], wts, Memr[kernel], nx*ny)
	    }
	    cnvwt = sqrt (wts)

	    if (in[2] == NULL)
		bpm2 = NULL
	    else
		bpm2 = bpm[2] 
	    if (bpm[1] == NULL && bpm2 == NULL)
		dobpm = false
	    else
		dobpm = true
	    if (dobpm) {
		call malloc (bpbuf, nc*ny, TY_REAL)
		call malloc (bpwts, nc, TY_REAL)
		call calloc (symwts, ny*ny+1, TY_POINTER)
		Memi[symwts+ny*ny] = -1
	    }

	    # Check for any line symmetries in the kernel.
	    call malloc (sym, ny, TY_INT)
	    call calloc (symbuf, ny*ny+1, TY_POINTER)
	    Memi[symbuf+ny*ny] = -1
	    do i = ny, 1, -1 {
		kptr = kernel + (i - 1) * nx
		do j = ny, 1, -1 {
		    ptr = kernel + (j - 1) * nx
		    do k = 0, nx-1 {
			if (Memr[kptr+k] != Memr[ptr+k])
			    break
		    }
		    if (k == nx) {
			Memi[sym+i-1] = j
			break
		    }
		}
	    }
	    do i = ny, 1, -1 {
		k = 0
		do j = ny, 1, -1
		    if (Memi[sym+j-1] == i)
			k = k + 1
		if (k == 1)
		    Memi[sym+i-1] = 0
	    }

	    call malloc (buf, nc*ny, TY_REAL)
	    call malloc (sbuf, nc*ny, TY_REAL)
	    call malloc (sigbuf, nc*ny, TY_REAL)
	    call malloc (ebuf, nc*ny, TY_REAL)
	    if (in[2] != NULL) {
		call malloc (buf2, nc*ny, TY_REAL)
		call malloc (buf3, nc*ny, TY_REAL)
		call malloc (sbuf2, nc*ny, TY_REAL)
		call malloc (sbuf3, nc*ny, TY_REAL)
		call malloc (sigbuf2, nc*ny, TY_REAL)
		call malloc (ebuf2, nc*ny, TY_REAL)
	    }

	    if (in[2] != NULL) {
		overlap = true
		if (1-offset[1] < 1 || nc-offset[1] > IM_LEN(in[2],1))
		    overlap = false
		if (1-offset[2] < 1 || nl-offset[2] > IM_LEN(in[2],2))
		    overlap = false
	    }
	    do i = 1, ny {
		call cnvgline1 (i, offset, in, bpm, indata, bp, bpval,
		    bpdet, bpflg)
		call cnvgline2 (i, offset, in, sky, sig, exp, skydata, sigdata,
		   expdata)
		off = mod (i, ny) * nc
		call amovr (Memr[indata[1]], Memr[buf+off], nc)
		call amovr (Memr[skydata[1]], Memr[sbuf+off], nc)
		call amovr (Memr[sigdata[1]], Memr[sigbuf+off], nc)
		call amovr (Memr[expdata[1]], Memr[ebuf+off], nc)
		if (in[2] != NULL) {
		    call amovr (Memr[indata[2]], Memr[buf2+off], nc)
		    call amovr (Memr[sigdata[2]], Memr[sigbuf2+off], nc)
		    call amovr (Memr[expdata[2]], Memr[ebuf2+off], nc)
		    call asubr_scale (Memr[buf+off], scale[1],
			Memr[buf2+off], scale[2], Memr[buf3+off], nc)
		}
		if (dobpm)
		    call amovi (Memi[bp], Memi[bpbuf+off], nc)
	    }
	}

	# Get new line.
	j = line +  ny2
	if (j > ny && j <= nl) {
	    call cnvgline1 (j, offset, in, bpm, indata, bp, bpval, bpdet, bpflg)
	    call cnvgline2 (j, offset, in, sky, sig, exp, skydata, sigdata,
	       expdata)
	    off = mod (j, ny) * nc
	    call amovr (Memr[indata[1]], Memr[buf+off], nc)
	    call amovr (Memr[skydata[1]], Memr[sbuf+off], nc)
	    call amovr (Memr[sigdata[1]], Memr[sigbuf+off], nc)
	    call amovr (Memr[expdata[1]], Memr[ebuf+off], nc)
	    if (in[2] != NULL) {
		call amovr (Memr[indata[2]], Memr[buf2+off], nc)
		call asubr_scale (Memr[buf+off], scale[1],
		    Memr[buf2+off], scale[2], Memr[buf3+off], nc)
		call amovr (Memr[skydata[2]], Memr[sbuf2+off], nc)
		call amovr (Memr[sigdata[2]], Memr[sigbuf2+off], nc)
		call amovr (Memr[expdata[2]], Memr[ebuf2+off], nc)
		call asubr_scale (Memr[sbuf+off], scale[1],
		    Memr[sbuf2+off], scale[2], Memr[sbuf3+off], nc)
	    }
	    if (dobpm) {
		ptr = bpbuf + off
		call amovi (Memi[bp], Memi[ptr], nc)
	    }
	}

	# Compute the convolution vector with boundary reflection.
	# Save and reuse lines with the same kernel weights apart
	# from a scale factor.

	kptr = kernel
	call aclrr (Memr[cnvdata], nc)
	call aclrr (Memr[scnvdata], nc)
	if (dobpm)
	    call aclrr (Memr[bpwts], nc)
	do i = 1, ny {
	    j = line + i - ny2 - 1
	    if (j < 1)
		j = 2 - j
	    else if (j > nl)
		j = 2 * nl - j
	    off = mod (j, ny) * nc
	    if (in[2] == NULL) {
		ptr = buf
		sptr = sbuf
	    } else {
		ptr = buf3
		sptr = sbuf3
	    }
	    k = Memi[sym+i-1]
	    if (k == 0) {
		mode = 1
		symptr = ptr
		symwptr = bpwts
	    } else {
		if (k == i)
		    mode = 2
		else
		    mode = 3
		symptr = Memi[symbuf+(k-1)*ny+mod(j,ny)]
		if (symptr == NULL) {
		    call malloc (symptr, nc, TY_REAL)
		    Memi[symbuf+(k-1)*ny+mod(j,ny)] = symptr
		    mode = 2
		}
		if (dobpm) {
		    symwptr = Memi[symwts+(k-1)*ny+mod(j,ny)]
		    if (symwptr == NULL) {
			call malloc (symwptr, nc, TY_REAL)
			Memi[symwts+(k-1)*ny+mod(j,ny)] = symwptr
		    }
		}
	    }
	    if (dobpm) {
		call convolve2 (Memr[ptr+off], Memr[cnvdata], Memr[symptr],
		    nc, Memr[kptr], Memr[scales+i-1], nx, Memi[bpbuf+off],
		    Memr[wtsl+i-1], Memr[bpwts], Memr[symwptr], mode)
		call convolve2 (Memr[sptr+off], Memr[scnvdata], Memr[symptr],
		    nc, Memr[kptr], Memr[scales+i-1], nx, Memi[bpbuf+off],
		    Memr[wtsl+i-1], Memr[bpwts], Memr[symwptr], mode)
	    } else {
		call convolve1 (Memr[ptr+off], Memr[cnvdata], Memr[symptr],
		    nc, Memr[kptr], Memr[scales+i-1], nx, mode)
		call convolve1 (Memr[sptr+off], Memr[scnvdata], Memr[symptr],
		    nc, Memr[kptr], Memr[scales+i-1], nx, mode)
	    }
	    kptr = kptr + nx
	}
	if (dobpm) {
	    do i = 0, nc-1 {
		if (Memr[bpwts+i] != 0.) {
		    Memr[cnvdata+i] = Memr[cnvdata+i] / Memr[bpwts+i]
		    Memr[scnvdata+i] = Memr[scnvdata+i] / Memr[bpwts+i]
		}
	    }
	}

	# Set the output vectors.
	off = mod (line, ny) * nc
	indata[1] = buf + off 
	skydata[1] = sbuf + off 
	sigdata[1] = sigbuf + off 
	expdata[1] = ebuf + off 
	if (dobpm) {
	    if (bpm2 == NULL)
		bp = bpbuf + off
	    else
		call amovi (Memi[bpbuf+off], Memi[bp], nc)
	}
	if (in[2] != NULL) {
	    if (overlap) {
		indata[2] = buf2 + off
		skydata[2] = sbuf2 + off
		sigdata[2] = sigbuf2 + off
		expdata[2] = ebuf2 + off
	    } else {
		call amovr (Memr[buf2+off], Memr[indata[2]], nc)
		call amovr (Memr[sbuf2+off], Memr[skydata[2]], nc)
		call amovr (Memr[sigbuf2+off], Memr[sigdata[2]], nc)
		call amovr (Memr[ebuf2+off], Memr[expdata[2]], nc)
	    }
	}
end
