include	<acecat.h>
include	<acecat1.h>


# CATWCS -- Set catalog WCS information.

procedure catwcs (cat, im)

pointer	cat			#I Catalog pointer
pointer	im			#I IMIO pointer

pointer	mw, hdr

pointer	mw_openim()
errchk	mw_openim

begin
	if (cat == NULL)
	    return

	# Set catalog WCS.
	mw = mw_openim (im)
	call catwcs_open (cat, mw)
	call mw_close (mw)

	# Set catalog header WCS.
	mw = CAT_MW(CAT_WCS(cat))
	hdr = CAT_OHDR(cat)
	call mw_saveim (mw, hdr)
end


# CATWCS_OPEN -- Open a WCS attached to a catalog.
#
# This copies the MWCS pointer so it is safe for the application to
# close it without affecting the catalog WCS.  Note that the WCS need
# not be that in the catalog header.

procedure catwcs_open (cat, mw)

pointer	cat			#I Catalog pointer
pointer	mw			#I MWCS pointer

pointer	mw_newcopy()

begin
	if (cat == NULL)
	    return

	call catwcs_close (cat)

	if (mw == NULL)
	    return

	call calloc (CAT_WCS(cat), CAT_WCSLEN, TY_STRUCT)
	CAT_MW(CAT_WCS(cat)) = mw_newcopy (mw)
end


# CATWCS_CLOSE -- Close a WCS attached to a catalog.
#
# This frees all memory associated with the WCS.  Note that we rely on
# mw_close to free any CT structures.

procedure catwcs_close (cat)

pointer	cat			#I Catalog pointer

begin
	if (cat == NULL)
	    return
	if (CAT_WCS(cat) == NULL)
	    return

	if (CAT_MW(CAT_WCS(cat)) != NULL)
	    call mw_close (CAT_MW(CAT_WCS(cat)))
	call mfree (CAT_WCS(cat), TY_STRUCT)
end


# CAT_LW -- Evaluate world coordinates from logical coordinates.
#
# For efficiency this routine does not error check if catalog WCS is defined.
# It also assumes the CT pointer is not initially set.

procedure cat_lw (cat, x, y, wx, wy, ra, dec)

pointer	cat			#I Catalog pointer
double	x			#I X logical coordinate
double	y			#I Y logical coordinate
double	wx			#I X world coordinate
double	wy			#I Y world coordinate
double	ra			#I RA world coordinate
double	dec			#I DEC world coordinate

char	axtype[3]
int	raaxis, decaxis
double	w[2]
pointer	mw, ct, mw_sctran()
bool	streq()

begin
	if (IS_INDEFD(x) || IS_INDEFD(y)) {
	    wx = INDEFD
	    wy = INDEFD
	    ra = INDEFD
	    dec = INDEFD
	    return
	}

	ct = CAT_CTLW(CAT_WCS(cat))
	if (ct == NULL) {
	    mw = CAT_MW(CAT_WCS(cat))
	    ct = mw_sctran (mw, "logical", "world", 03B)
	    CAT_CTLW(CAT_WCS(cat)) = ct
	    raaxis = INDEFI
	    decaxis = INDEFI
	    ifnoerr (call mw_gwattrs (mw, 1, "axtype", axtype, 3)) {
	        if (streq (axtype, "ra"))
		    raaxis = 1
		else if (streq (axtype, "dec"))
		    decaxis = 1
	    }
	    ifnoerr (call mw_gwattrs (mw, 2, "axtype", axtype, 3)) {
	        if (streq (axtype, "ra"))
		    raaxis = 2
		else if (streq (axtype, "dec"))
		    decaxis = 2
	    }
	}

	call mw_c2trand (ct, x, y, w[1], w[2])
	wx = w[1]
	wy = w[2]
	if (IS_INDEFI(raaxis))
	    ra = INDEFD
	else
	    ra = w[raaxis]
	if (IS_INDEFI(decaxis))
	    dec = INDEFD
	else
	    dec = w[decaxis]
end


# CAT_LP -- Evaluate physical coordinates from logical coordinates.
#
# For efficiency this routine does not error check if catalog WCS is defined.

procedure cat_lp (cat, x, y, px, py)

pointer	cat			#I Catalog pointer
double	x			#I X logical coordinate
double	y			#I Y logical coordinate
double	px			#I X physical coordinate
double	py			#I Y physical coordinate

pointer	ct, mw_sctran()

begin
	if (IS_INDEFD(x) || IS_INDEFD(y)) {
	    px = INDEFD
	    py = INDEFD
	    return
	}

	ct = CAT_CTLP(CAT_WCS(cat))
	if (ct == NULL) {
	    ct = mw_sctran (CAT_MW(CAT_WCS(cat)), "logical", "physical", 03B)
	    CAT_CTLP(CAT_WCS(cat)) = ct
	}

	call mw_c2trand (ct, x, y, px, py)
end
